import numpy as np
import itertools
from copy import deepcopy

# Using the generator pattern (an iterable)
class lattice_constant_generator:

	def __init__(self, lattice_constant_parameters):
		self.rounding_to = 8
		if isinstance(lattice_constant_parameters,tuple) or isinstance(lattice_constant_parameters,list):
			lattice_constant_parameters = {'c': lattice_constant_parameters}
		generators = []
		for lattice_constant_type, lattice_constant_parameter in lattice_constant_parameters.items():
			if (isinstance(lattice_constant_parameter,tuple) or isinstance(lattice_constant_parameter,list)) and len(lattice_constant_parameter) == 3:
				lc_low, lc_high, lc_resolution = lattice_constant_parameter
				if isinstance(lc_resolution,float):
					generators.append((lattice_constant_type, tuple(np.arange(lc_low, lc_high+lc_resolution, lc_resolution))))
				elif isinstance(lc_resolution,int):
					generators.append((lattice_constant_type, tuple(np.linspace(lc_low, lc_high, lc_resolution, endpoint=True))))
			else:
				generators.append((lattice_constant_type, lattice_constant_parameter))
		generators.sort()
		self.lattice_constant_types = sorted([lattice_constant_type for lattice_constant_type, generator in generators])
		self.generators = [generator for lattice_constant_type, generator in generators]
		self.lattice_constant_generator = itertools.product(*deepcopy(self.generators))

	def __iter__(self):
		return self

	def get_lattice_constant_types(self):
		return self.lattice_constant_types

	def reset(self):
		self.lattice_constant_generator = itertools.product(*deepcopy(self.generators))

	# Python 3 compatibility
	def __next__(self):
		return self.next()

	def next(self):
		try:
			lattice_constants = next(self.lattice_constant_generator)
			if isinstance(lattice_constants,tuple) or isinstance(lattice_constants,list):
				if len(lattice_constants) == 1:
					lattice_constants = round(lattice_constants[0],self.rounding_to)
				else:
					lattice_constants = {key: round(value,self.rounding_to) for key, value in zip(self.lattice_constant_types,lattice_constants)}
				#lattice_constants = {key: round(value,self.rounding_to) for key, value in zip(self.lattice_constant_types,lattice_constants)}
			return lattice_constants
		except StopIteration:
			raise StopIteration()


import numpy as np
import matplotlib.pyplot as plt
from matplotlib.lines import Line2D

def get_plotting_lims(lattice_point,percent_diff):
		lattice_point_diff = max(lattice_point) - min(lattice_point)
		lattice_point_diff_with_percent  = lattice_point_diff*(percent_diff/100.0)
		lattice_point_low  = min(lattice_point) - lattice_point_diff_with_percent
		lattice_point_high = max(lattice_point) + lattice_point_diff_with_percent
		return lattice_point_low, lattice_point_high

def plot_energy_vs_lattice_constants_1D(lattice_energies_dict, limits, plotting_limits, minimum_energy, lowest_energy_lattice_constants):
	lattice_point_c = []; lattice_energies = [];
	lattice_point_c_bottom, lattice_point_c_top = limits['c']
	for cc_LC, energy_per_atom in sorted(list(lattice_energies_dict.items())):
		if (lattice_point_c_bottom <= cc_LC <= lattice_point_c_top):
			lattice_point_c.append(cc_LC)
			lattice_energies.append(energy_per_atom)
	plt.scatter(lattice_point_c, lattice_energies, s=5, alpha=0.5,zorder=10)
	for cc_point in lowest_energy_lattice_constants:
		label = 'c = '+str(cc_point)+' '+r'$\AA$' #for aa_point, cc_point in lowest_energy_lattice_constants
		plt.scatter(lowest_energy_lattice_constants, minimum_energy, color='r', s=5, alpha=0.5,zorder=20, label=label)
	plt.xlabel('Lattice Constant ('+r'$\AA$'+')')
	plt.ylabel('Energy per Atom ('+r'$eV/Atom$'+')')
	percent_diff = 2.0
	lattice_point_low, lattice_point_high = get_plotting_lims(lattice_point_c,percent_diff)
	lattice_energies_low, lattice_energies_high = get_plotting_lims(lattice_energies,percent_diff)
	#x_diff = lattice_energies_high - lattice_energies_low
	#plt.hlines(minimum_energy,lattice_energies_low - x_diff,lattice_energies_high + x_diff,color='k',linestyles='dashed',zorder=1)
	plt.xlim((lattice_point_low,lattice_point_high))
	plt.ylim((lattice_energies_low,lattice_energies_high))
	leg = plt.legend(title='Cohensive Energy: '+str(round(minimum_energy,5))+' '+r'$eV/Atom$')
	plt.savefig('Energy_Vs_Lattice_Constant.png')
	plt.savefig('Energy_Vs_Lattice_Constant.svg')
	plt.savefig('Energy_Vs_Lattice_Constant.eps')
	plt.cla(); plt.clf()

def plot_energy_vs_lattice_constants_2D(lattice_energies_dict, limits, plotting_limits, minimum_energy, lowest_energy_lattice_constants):
	lattice_point_a = []; lattice_point_c = []; lattice_energies = [];
	lattice_point_a_bottom, lattice_point_a_top = limits['a']
	lattice_point_c_bottom, lattice_point_c_top = limits['c']
	for (aa_LC, cc_LC), energy_per_atom in sorted(list(lattice_energies_dict.items())):
		if (lattice_point_a_bottom <= aa_LC <= lattice_point_a_top) and (lattice_point_c_bottom <= cc_LC <= lattice_point_c_top):
			lattice_point_a.append(aa_LC)
			lattice_point_c.append(cc_LC)
			lattice_energies.append(energy_per_atom)
	lattice_point_a  = np.array(lattice_point_a)
	lattice_point_c  = np.array(lattice_point_c)
	lattice_energies =  np.array(lattice_energies)

	zero_zero_value = lattice_point_a[0]
	for row_length in range(len(lattice_point_a)):
		if not lattice_point_a[row_length] == zero_zero_value:
			break
	column_length = int(float(len(lattice_point_a))/float(row_length))

	lattice_point_a_matrix = lattice_point_a.reshape(column_length,row_length)
	lattice_point_c_matrix = lattice_point_c.reshape(column_length,row_length)

	def Z_function(AA,CC):
		row_total_index, column_total_index = AA.shape
		data = np.zeros(shape=(row_total_index, column_total_index))#,fill_value=None)
		for row_index in range(row_total_index):
			for column_index in range(column_total_index):
				aa_LC = AA[row_index][column_index]
				cc_LC = CC[row_index][column_index]
				data[row_index][column_index] = lattice_energies_dict[(aa_LC,cc_LC)]
		#data = np.real(data)
		return data
	lattice_energies_matrix = Z_function(lattice_point_a_matrix, lattice_point_c_matrix)

	from mpl_toolkits import mplot3d
	fig = plt.figure()
	ax = plt.axes(projection='3d')

	levels_conf = np.linspace(min(lattice_energies), max(lattice_energies), 20)
	levels_confull = np.linspace(min(lattice_energies), max(lattice_energies), 1000)

	# https://jakevdp.github.io/PythonDataScienceHandbook/04.12-three-dimensional-plotting.html
	ax.plot_surface(lattice_point_a_matrix, lattice_point_c_matrix, lattice_energies_matrix, rstride=1, cstride=1,cmap='viridis', edgecolor='none', alpha=0.2, zorder=10)
	ax.contour3D(lattice_point_a_matrix, lattice_point_c_matrix, lattice_energies_matrix, levels=levels_conf, cmap='viridis', zorder=11)
	ax.set_xlabel('Lattice Constant (a) ('+r'$\AA$'+')')
	ax.set_ylabel('Lattice Constant (c) ('+r'$\AA$'+')')
	ax.set_zlabel('Cohensive Energy ('+r'$eV/Atom$'+')')
	percent_diff = 2.0
	lattice_point_a_low, lattice_point_a_high = get_plotting_lims(lattice_point_a,0)
	lattice_point_c_low, lattice_point_c_high = get_plotting_lims(lattice_point_c,0)
	_, lattice_energies_high = get_plotting_lims(lattice_energies,0)
	lattice_energies_low, _  = get_plotting_lims(lattice_energies,50)
	ax.set_xlim((lattice_point_a_low,lattice_point_a_high))
	ax.set_ylim((lattice_point_c_low,lattice_point_c_high))
	ax.set_zlim((lattice_energies_low,lattice_energies_high))
	# https://stackoverflow.com/questions/35445424/surface-and-3d-contour-in-matplotlib
	ax.contour(lattice_point_a_matrix, lattice_point_c_matrix, lattice_energies_matrix, levels=levels_conf, cmap='viridis', linestyles="solid",offset=lattice_energies_low,zorder=-11)
	plt.contourf(lattice_point_a_matrix, lattice_point_c_matrix, lattice_energies_matrix, levels=levels_confull, cmap='viridis', alpha=0.2, antialiased=True, linestyles=None,offset=lattice_energies_low,zorder=-10)
	
	lowest_lattice_points_a = []
	lowest_lattice_points_c = []
	for aa_point, cc_point in lowest_energy_lattice_constants:
		lowest_lattice_points_a.append(aa_point)
		lowest_lattice_points_c.append(cc_point)
		label = '(a = '+str(aa_point)+' '+r'$\AA$'+', c = '+str(cc_point)+' '+r'$\AA$'+')' #for aa_point, cc_point in lowest_energy_lattice_constants
		ax.scatter(lowest_lattice_points_a,lowest_lattice_points_c,[lattice_energies_low]*len(lowest_lattice_points_a),color='red',zorder=40, label=label)
		#ax.contour(lattice_point_a_matrix, lattice_point_c_matrix, lattice_energies_matrix, 10, lw=3, colors="k", linestyles="solid")
	leg = ax.legend(title='Cohensive Energy: '+str(round(minimum_energy,5))+' eV/Atom', facecolor='white', framealpha=0.8)#,handlelength=0, handletextpad=0, markerscale=0)
	#leg.set_facecolor('white')
	
	ax.scatter(lowest_lattice_points_a,lowest_lattice_points_c,[minimum_energy]*len(lowest_lattice_points_a),color='red',zorder=-9)

	cbar = plt.colorbar(pad=0.12)
	cbar.set_label('Cohensive Energy ('+r'$eV/Atom$'+')', rotation=270, labelpad=20)

	plt.tight_layout()
	plt.savefig('Energy_Vs_Lattice_Constant.png')
	plt.savefig('Energy_Vs_Lattice_Constant.svg')
	plt.savefig('Energy_Vs_Lattice_Constant.eps')
	plt.cla(); plt.clf()
	#plt.show()

	#https://jakevdp.github.io/PythonDataScienceHandbook/04.04-density-and-contour-plots.html
	fig, ax = plt.subplots()
	#ax.view_init(azim=0, elev=90)
	
	contours = plt.contour(lattice_point_a_matrix, lattice_point_c_matrix, lattice_energies_matrix, levels=levels_conf, cmap='viridis', zorder=10)
	plt.clabel(contours, inline=True, fontsize=10)

	levels_confull = np.linspace(min(lattice_energies), max(lattice_energies), 1000)
	plt.contourf(lattice_point_a_matrix, lattice_point_c_matrix, lattice_energies_matrix, levels=levels_confull, cmap='viridis', alpha=0.2, antialiased=True, linestyles=None, zorder=20)
	#plt.imshow(lattice_energies_matrix, extent=[lattice_point_a_low, lattice_point_a_high, lattice_point_c_low, lattice_point_c_high], origin='lower',cmap='viridis', alpha=0.2) # , interpolation='bilinear'
	
	#import pdb; pdb.set_trace()
	for aa_point, cc_point in lowest_energy_lattice_constants:
		label = '(a = '+str(aa_point)+' '+r'$\AA$'+', c = '+str(cc_point)+' '+r'$\AA$'+')' #for aa_point, cc_point in lowest_energy_lattice_constants
		plt.plot(lowest_lattice_points_a, lowest_lattice_points_c, 'ro', label=label, zorder=30)
	leg = ax.legend(title='Cohensive Energy: '+str(round(minimum_energy,5))+' '+r'$eV/Atom$', facecolor='white', framealpha=0.8) #,handlelength=0, handletextpad=0, markerscale=0)
	#leg.set_facecolor('white')
	#for item in leg.legendHandles:
	#	item.set_visible(False)

	cbar = plt.colorbar()
	cbar.set_label('Cohensive Energy ('+r'$ev/Atom$'+')', rotation=270, labelpad=20)
	#lattice_point_a_low, lattice_point_a_high = get_plotting_lims(lattice_point_a,0)
	#lattice_point_c_low, lattice_point_c_high = get_plotting_lims(lattice_point_c,0)
	#lattice_energies_low, lattice_energies_high = get_plotting_lims(lattice_energies,0)
	#import pdb; pdb.set_trace()

	plt.xlim((lattice_point_a_low,lattice_point_a_high))
	plt.ylim((lattice_point_c_low,lattice_point_c_high))
	plt.xlabel('Lattice Constant (a) ('+r'$\AA$'+')')
	plt.ylabel('Lattice Constant (c) ('+r'$\AA$'+')')

	plt.tight_layout()
	#plt.zlim((lattice_energies_low,lattice_energies_high))
	plt.savefig('Energy_Vs_Lattice_Constant_Contour.png')
	plt.savefig('Energy_Vs_Lattice_Constant_Contour.svg')
	plt.savefig('Energy_Vs_Lattice_Constant_Contour.eps')
	plt.cla(); plt.clf()


import multiprocessing as mp

def get_energies_across_lattice_constants_ASE(lattice_type,symbol,lattice_constant_generator,lattice_constant_types,size,directions=None,miller=None,calculator=None,no_of_cpus=1,lattice_data_file=None,energies_vs_lattice_constants={}):
	"""
	This method allows the user to obtain lattice constant information about your system using an ase based calculator (and local optimiser)
	"""
	print('================================================================================')
	global calculator_ASE
	calculator_ASE = calculator
	if no_of_cpus == 1:
		energies_vs_lattice_constants = get_energies_across_lattice_constants_ASE_one_cpu(lattice_type,symbol,lattice_constant_generator,lattice_constant_types,size,directions,miller,lattice_data_file,energies_vs_lattice_constants)
	else:
		energies_vs_lattice_constants = get_energies_across_lattice_constants_ASE_multi_cpu(lattice_type,symbol,lattice_constant_generator,lattice_constant_types,size,directions,miller,no_of_cpus,lattice_data_file,energies_vs_lattice_constants)
	print('================================================================================')
	return energies_vs_lattice_constants

def get_energies_across_lattice_constants_ASE_one_cpu(lattice_type,symbol,lattice_constant_generator,lattice_constant_types,size,directions=None,miller=None,lattice_data_file=None,energies_vs_lattice_constants={}):
	"""

	"""
	for latticeconstants in lattice_constant_generator:
		get_energies_across_lattice_constants_ASE_single(lattice_type,symbol,latticeconstants,lattice_constant_types,size,directions,miller,lattice_data_file,energies_vs_lattice_constants,qq=None)
	return energies_vs_lattice_constants

def get_energies_across_lattice_constants_ASE_single(lattice_type,symbol,latticeconstants,lattice_constant_types,size,directions,miller,lattice_data_file,energies_vs_lattice_constants,qq=None):
	"""

	"""
	if isinstance(latticeconstants,dict):
		latticeconstants_for_dict = tuple(latticeconstants[lattice_constant_types[index]] for index in range(len(lattice_constant_types)))
		one_lattice_constant = False
	else:
		latticeconstants_for_dict = latticeconstants
		one_lattice_constant = True
	if latticeconstants_for_dict in energies_vs_lattice_constants.keys():
		return
	print(latticeconstants)
	bulk_system = lattice_type(symbol=symbol, latticeconstant=latticeconstants, size=size) #, directions=directions, miller=miller)
	global calculator_ASE

	bulk_system.set_calculator(calculator_ASE)
	energy_per_atom = get_energy_per_atom(bulk_system)
	if one_lattice_constant:
		volume_per_atom = get_volume_per_atom(bulk_system)
		energies_vs_lattice_constants[latticeconstants_for_dict] = (energy_per_atom,volume_per_atom)
	else:
		volume_per_atom = None
		energies_vs_lattice_constants[latticeconstants_for_dict] = energy_per_atom
	if not qq is None:
		res = (lattice_data_file,latticeconstants_for_dict,energy_per_atom,volume_per_atom)
		qq.put(res)
	else:
		save_datum_to_file(lattice_data_file,latticeconstants_for_dict,energy_per_atom,volume_per_atom)

def get_energy_per_atom(bulk_system):
	energy = bulk_system.get_potential_energy(bulk_system)
	energy_per_atom = energy/float(len(bulk_system))
	return energy_per_atom

def get_volume_per_atom(bulk_system):
	volume_per_atom = round(bulk_system.get_volume()/float(len(bulk_system)),9)
	return volume_per_atom

def save_datum_to_file(lattice_data_file,latticeconstants,energy_per_atom, volume=None):
	"""

	"""
	with open(lattice_data_file,'a') as lattice_data_FILE:
		if volume == None:
			lattice_data_FILE.write(str(latticeconstants)+': '+str(energy_per_atom)+'\n')
		else:
			lattice_data_FILE.write(str(latticeconstants)+': '+str(energy_per_atom)+' ('+str(volume)+')\n')

def get_system_from_ASE(lattice_type,symbol,latticeconstants,size,directions,miller,calculator):
	"""

	"""	
	bulk_systems = []
	for a_latticeconstants in latticeconstants:
		import pdb; pdb.set_trace()
		bulk_system = lattice_type(symbol=symbol, latticeconstant=a_latticeconstants, size=size)
		bulk_system.set_calculator(calculator)
		bulk_systems.append(bulk_system)
	return bulk_systems

# --------------------------------------------------------------------------------------------------------------------------------------------------------------------------------v
import multiprocessing as mp
def get_energies_across_lattice_constants_ASE_multi_cpu(lattice_type,symbol,lattice_constant_generator,lattice_constant_types,size,directions=None,miller=None,no_of_cpus=1,lattice_data_file=None,energies_vs_lattice_constants={}):
	"""

	"""
	# solution from https://stackoverflow.com/questions/13446445/python-multiprocessing-safely-writing-to-a-file
	manager = mp.Manager()
	qq = manager.Queue()    
	pool = mp.Pool(processes=no_of_cpus)
	#put listener to work first
	watcher = pool.apply_async(listener, (qq,))
	#fire off workers
	dictionary_multiprocessing = manager.dict()
	dictionary_multiprocessing.update(energies_vs_lattice_constants)
	del energies_vs_lattice_constants
	print('Performing calculations upon Lattices')
	jobs = []
	for latticeconstants in lattice_constant_generator:
		task = (lattice_type,symbol,latticeconstants,lattice_constant_types,size,directions,miller,lattice_data_file,dictionary_multiprocessing,qq)
		job = pool.apply_async(get_energies_across_lattice_constants_ASE_single, task)
		jobs.append(job)
	# collect results from the workers through the pool result queue
	for job in jobs: 
		job.get()
	print('Finished performing calculations upon Lattices')
	print('Performing last finishing off pieces of work')
	energies_vs_lattice_constants = dict(dictionary_multiprocessing)
	#now we are done, kill the listener
	qq.put('kill')
	pool.close()
	pool.join()
	print('Finished performing last finishing off pieces of work')
	return energies_vs_lattice_constants

def listener(qq):
	'''listens for messages on the q, writes to file. '''	
	while True:
		mm = qq.get()
		if mm == 'kill':
			break
		lattice_data_file,latticeconstants_for_dict,energy_per_atom,volume = mm
		save_datum_to_file(lattice_data_file,latticeconstants_for_dict,energy_per_atom,volume)




#!/usr/bin/env python3
'''
Geoffrey Weal, LatticeFinder_Tidy_Finished_Jobs.py, 30/04/2021

This program is designed to remove the files of jobs that have finished. 

'''
import os
from ase.io import read

files_to_remove = ['CHG','CHGCAR','CONTCAR','DOSCAR','EIGENVAL','IBZKPT','INCAR','KPOINTS','OSZICAR','PCDAT','POSCAR','POTCAR','REPORT','submit.sl','vasprun.xml','WAVECAR','XDATCAR']

for root, dirs, files in os.walk("."):
    dirs.sort()
    files.sort()
    if 'OUTCAR' in files:
        try:
            system = read(root+'/'+'OUTCAR')
            system.get_potential_energy()
            system.get_volume()
        except:
            dirs[:] = []
            files[:] = []
            continue
        print(root)
        for file_to_remove in files_to_remove:
            if file_to_remove in files:
                os.remove(root+'/'+file_to_remove)
        for file in files:
            if file.startswith('slurm-') and file.endswith('.out'):
                os.remove(root+'/'+file)
            elif file.startswith('slurm-') and file.endswith('.err'):
                os.remove(root+'/'+file)
        dirs[:] = []
        files[:] = []


'''
Geoffrey Weal, Create_submitSL_slurm_Main.py, 10/02/2021

This program is designed to create the various forms of submit.sl/mass_submit.sl files that could be used to submit genetic algorithm trials to slurm.
'''

def make_submitSL(local_path,project,time,nodes,ntasks_per_node,mem_per_cpu,partition='large',email='',python_version='Python/3.6.3-gimkl-2017a',vasp_version='VASP/5.4.4-intel-2017a',vasp_execution='vasp_std'):
    # create name for job
    print("creating submit.sl for "+str(local_path))
    name = 'Getting_LatticeFinder_Data_'+local_path.replace('/','_')
    # writing the submit.sl script
    with open(local_path+"/submit.sl", "w") as submitSL:
        submitSL.write('#!/bin/bash -e\n')
        submitSL.write('#SBATCH -J ' + str(name) + '\n')
        submitSL.write('#SBATCH -A ' + str(project) + '         # Project Account\n')
        submitSL.write('#SBATCH --partition ' + str(partition) + '\n')
        submitSL.write('\n')
        submitSL.write('#SBATCH --time=' + str(time) + '     # Walltime\n')
        submitSL.write('#SBATCH --nodes=' + str(nodes) + '\n')
        submitSL.write('#On VASP, Ben Roberts recommends using the same number\n')
        submitSL.write('#of tasks on all nodes, even if this makes scheduling\n')
        submitSL.write('#a little more difficult\n')
        submitSL.write('#SBATCH --ntasks-per-node=' + str(ntasks_per_node) + '\n')
        submitSL.write('#SBATCH --mem-per-cpu=' + str(mem_per_cpu) + '\n')
        #submitSL.write('#SBATCH -C sb\n')
        submitSL.write('\n')
        #submitSL.write("#SBATCH --hint=nomultithread    # don't use hyperthreading"+'\n')
        submitSL.write('#SBATCH --output=slurm-%j.out      # %x and %j are replaced by job name and ID'+'\n')
        submitSL.write('#SBATCH --error=slurm-%j.err'+'\n')
        if not email == '':
            submitSL.write('#SBATCH --mail-user=' + str(email) + '\n')
            submitSL.write('#SBATCH --mail-type=ALL\n')
        submitSL.write('#SBATCH --hint nomultithread\n')
        submitSL.write('\n')
        submitSL.write('module load '+str(vasp_version)+'\n')
        submitSL.write('srun -K '+str(vasp_execution)+'\n')
        submitSL.write('\n')
        submitSL.write('# removing files except for OUTCAR as we assume it finished successfully\n')
        submitSL.write('module load '+str(python_version)+'\n')
        submitSL.write('Check_LatticeFinder_to_Tidy_after_Job.py')

def make_submitSL_packets_for_latticeFinder(number_of_vasp_calc_to_run_per_packet,directories,local_path,project,newlist,time,nodes,ntasks_per_node,mem_per_cpu,partition='large',email='',python_version='Python/3.6.3-gimkl-2017a',vasp_version='VASP/5.4.4-intel-2017a',vasp_execution='vasp_std'):
    number_of_vasp_calc_to_run_per_packet = 25
    mass_submit_counter = 1
    while len(directories) > 0:
        if len(directories) > nn:
            newlist = directories[:nn]
            del directories[:nn]
        else:
            newlist = list(directories)
            directories = []
        print(newlist)
        make_submitSL_packets_for_latticeFinder_subsidiary_method(mass_submit_counter,local_path,project,newlist,time,nodes,ntasks_per_node,mem_per_cpu,partition=partition,email=email,python_version=python_version,vasp_version=vasp_version,vasp_execution=vasp_execution)
        mass_submit_counter += 1

def make_submitSL_packets_for_latticeFinder_subsidiary_method(mass_submit_counter,local_path,project,newlist,time,nodes,ntasks_per_node,mem_per_cpu,partition='large',email='',python_version='Python/3.6.3-gimkl-2017a',vasp_version='VASP/5.4.4-intel-2017a',vasp_execution='vasp_std'):
    # create name for job
    print("creating collective_LatticeFinder_submit_"+str(mass_submit_counter)+".sl for "+str(local_path))
    name = local_path.replace('/','_')
    # writing the mass_submit.sl script
    with open(local_path+'/'+"collective_LatticeFinder_submit_"+str(mass_submit_counter)+".sl", "w") as submitSL:
        submitSL.write('#!/bin/bash -e\n')
        submitSL.write('#SBATCH -J ' + str(name) + '\n')
        submitSL.write('#SBATCH -A ' + str(project) + '         # Project Account\n')
        submitSL.write('\n')
        submitSL.write('#SBATCH --time=' + str(time) + '     # Walltime\n')
        submitSL.write('#SBATCH --nodes=' + str(nodes) + '\n')
        submitSL.write('#SBATCH --ntasks-per-node=' + str(ntasks_per_node) + '\n')
        submitSL.write('#SBATCH --mem-per-cpu=' + str(mem_per_cpu) + '\n')
        #submitSL.write('#SBATCH -C sb\n')
        submitSL.write('\n')
        submitSL.write('#SBATCH --partition='+str(partition)+'\n')
        submitSL.write('#SBATCH --output=slurm-%j.out      # %x and %j are replaced by job name and ID'+'\n')
        submitSL.write('#SBATCH --error=slurm-%j.err'+'\n')
        if not email == '':
            submitSL.write('#SBATCH --mail-user=' + str(email) + '\n')
            submitSL.write('#SBATCH --mail-type=ALL\n')
        #submitSL.write('\n')
        submitSL.write('#SBATCH --hint nomultithread\n')
        submitSL.write('\n')
        submitSL.write('######################\n')
        submitSL.write('# Begin work section #\n')
        submitSL.write('######################\n')
        submitSL.write('\n')
        submitSL.write('module load '+str(python_version)+'\n')
        submitSL.write('module load '+str(vasp_version)+'\n')
        submitSL.write('\n')
        submitSL.write('my_array=('+' '.join(newlist)+')\n')
        submitSL.write('\n')
        submitSL.write('for directory in "${my_array[@]}"\n')
        submitSL.write('do\n')
        submitSL.write('    echo "======================================"\n')
        submitSL.write('    echo "$directory"\n')
        submitSL.write('    cd "$directory"\n')
        submitSL.write('    srun -K '+str(vasp_execution)+'\n')
        submitSL.write('    # removing files except for OUTCAR as we assume it finished successfully\n')
        submitSL.write('    LatticeFinder_Tidy_after_Job.py\n')
        submitSL.write('    cd ..\n')
        submitSL.write('    echo "======================================"\n')
        submitSL.write('done\n')


from copy import deepcopy

def round_lattice_constant_parameters(lattice_constant_parameters):
	rounding_to = 8
	lattice_constant_parameters_copy = deepcopy(lattice_constant_parameters)
	new_lattice_constant_parameters = []
	if isinstance(lattice_constant_parameters_copy,list) or isinstance(lattice_constant_parameters_copy,tuple):
		for number in lattice_constant_parameters_copy:
			new_lattice_constant_parameters.append(round(number,rounding_to))
	return new_lattice_constant_parameters


import os
from ase.io import write

def get_energies_across_lattice_constants_in_Manual_Mode(lattice_type,symbol,lattice_constant_generator,size,directions=None,miller=None,folder_name='Manual_Mode_Clusters'):
	"""

	"""
	for latticeconstants in lattice_constant_generator:
		bulk_system = lattice_type(symbol=symbol, latticeconstant=latticeconstants, size=size) #, directions=directions, miller=miller)

def save_cluster_to_folder(folder,name,filename_suffix,manual_mode,cluster):
	path_to_folder = folder+'/'+name
	if not os.path.exists(path_to_folder):
		os.mkdir(path_to_folder)
	if manual_mode == 'vasp':
		write(path_to_folder+'/'+'POSCAR',cluster,format='vasp')
	else:
		write(path_to_folder+'/'+name+'.'+filename_suffix,cluster,format='xyz')

def get_system_from_Manual_Mode(latticeconstants,Manual_Mode_Cluster_Folder):
	pass


import os, time
import numpy as np
from shutil import copyfile
from ase.io import write
import multiprocessing as mp
import matplotlib.pyplot as plt

from ase.calculators.calculator import Calculator

from LatticeFinder.LatticeFinder.lattice_constant_generator import lattice_constant_generator

from ase.lattice.cubic import SimpleCubic, FaceCenteredCubic, BodyCenteredCubic, Diamond
from ase.lattice.tetragonal import SimpleTetragonal, CenteredTetragonal
from ase.lattice.orthorhombic import SimpleOrthorhombic, BaseCenteredOrthorhombic, FaceCenteredOrthorhombic, BodyCenteredOrthorhombic
from ase.lattice.monoclinic import SimpleMonoclinic, BaseCenteredMonoclinic
from ase.lattice.triclinic import Triclinic
from ase.lattice.hexagonal import Hexagonal, HexagonalClosedPacked, Graphite
from ase.lattice.compounds import B1, B2, B3, L1_2, L1_0

from LatticeFinder.LatticeFinder.examining_lattice_constant_methods_with_ASE         import get_energies_across_lattice_constants_ASE, get_system_from_ASE
from LatticeFinder.LatticeFinder.examining_lattice_constant_methods_with_VASP        import get_energies_across_lattice_constants_VASP, get_system_from_VASP
from LatticeFinder.LatticeFinder.examining_lattice_constant_methods_with_Manual_Mode import get_energies_across_lattice_constants_in_Manual_Mode, get_system_from_Manual_Mode

from LatticeFinder.LatticeFinder.plotting_methods import plot_energy_vs_lattice_constants_1D, plot_energy_vs_lattice_constants_2D

from LatticeFinder.LatticeFinder.other_methods import round_lattice_constant_parameters

class LatticeFinder_Program:
	"""
	This program is designed to give plots and data to help obtain the optimal lattice constants for a 2D and 3D system.
	"""
	def __init__(self, symbol, lattice_type, lattice_constant_parameters, calculator, size=(1,1,1), directions=None, miller=None, limits=None, no_of_cpus=1, slurm_information={}):
		# input options required for obtaining lattice constants
		self.check_input_information(symbol, lattice_type, lattice_constant_parameters, calculator, size, directions, miller, slurm_information)
		# Information about how to make plots
		self.check_output_information(limits)
		# Other options
		self.no_of_cpus = no_of_cpus
		# show initial information
		self.show_initial_messages(symbol, lattice_type, lattice_constant_parameters, calculator, size, directions, miller, limits, no_of_cpus, slurm_information)
		# Run the program
		self.run()

	def show_initial_messages(self, symbol, lattice_type, lattice_constant_parameters, calculator, size, directions, miller, limits, no_of_cpus, slurm_information):
		from LatticeFinder.__init__ import __version__
		print('=========================================================')
		print()
		print('            This is the LATTICEFINDER Program            ')
		print('            Version: '+str(__version__))
		print()
		print('=========================================================')
		print('Input settings:')
		print()
		print('symbol: '+str(symbol))
		print('lattice_type: '+str(lattice_type))
		print('lattice_constant_parameters: '+str(round_lattice_constant_parameters(self.lattice_constant_parameters)))
		print('calculator: '+str(calculator))
		print('size: '+str(size))
		print('directions: '+str(directions))
		print('miller: '+str(miller))
		print('=========================================================')
		print('Output settings:')
		print()
		print('limits')
		if self.calculator == 'vasp':
			print('slurm_information: '+str(slurm_information))
		print('=========================================================')
		print('Other settings:')
		print()
		print('no_of_cpus: '+str(no_of_cpus))
		print('=========================================================')

	def check_input_information(self, symbol, lattice_type, lattice_constant_parameters, calculator, size, directions, miller, slurm_information={}):
		"""
		This method will take the input data from input_information, set it ready for running in this program, and do some checking of the varables in input_information to make sure that are in a correct format.
		"""
		# General options
		self.symbol = symbol
		self.lattice_type_name = str(lattice_type)
		self.lattice_type = eval(lattice_type)
		self.lattice_constant_parameters = lattice_constant_parameters
		self.calculator = calculator # This could be a string called 'VASP' or 'Manual Mode'
		if isinstance(self.calculator,str):
			self.calculator = self.calculator.lower()
			if self.calculator == 'vasp':
				self.VASP_Input_files = 'VASP_Files'
				self.VASP_Systems_Folder = 'VASP_Systems'
				self.slurm_information = slurm_information
				self.check_and_note_slurm_information()
			elif self.calculator == 'manual mode':
				self.Manual_Mode_Cluster_Folder = 'Manual_Mode_Clusters'
			else:
				exit('Error, calculator needs to be a ASE calculator, "VASP" or "Manual Mode".')
		self.size = size
		# Other options about constructing the cell
		self.directions = directions
		self.miller = miller
		# The file of where to save lattice constant data to
		self.lattice_data_file = 'lattice_data.txt'

	def check_and_note_slurm_information(self):
		entry = 'Make individual or packet submitSL files'
		if entry in self.slurm_information:
			self.make_packets = self.slurm_information[entry]
		else:
			self.make_packets = 'packet'

	def check_output_information(self, limits):
		"""
		This method will check the variables for making output data such as plots.
		"""
		# Plotting variables
		if limits == None:
			if isinstance(limits,dict):
				self.limits = {lc_key: None for lc_key in self.lattice_constant_parameters.keys()}
			else:
				self.limits = {lc_key: None for lc_key in ['c']}
		elif isinstance(limits,dict):
			self.limits = limits
		else:
			self.limits = {lc_key: limits for lc_key in self.lattice_constant_parameters.keys()}
		self.plotting_limits = dict(self.limits)
		for key, value in self.limits.items():
			if value is None:
				self.limits[key] = (-float('inf'), float('inf'))
			elif isinstance(value,tuple) or isinstance(value,list):
				value = list(value)
				if value[0] is None:
					value[0] = -float('inf')
				if value[1] is None:
					value[1] = float('inf')
				self.limits[key] = tuple(value)

	############################################################################################################################

	def run(self):
		"""
		This method will run the LatticeFinder program commands. 
		"""
		# Perform checks on the lattice_constant_parameters to make sure it is in an acceptable format
		self.lattice_constant_parameters_checks(self.lattice_constant_parameters)
		# Set up the lattice constant generator and types names for future work
		self.lattice_constant_generator = lattice_constant_generator(self.lattice_constant_parameters)
		self.lattice_constant_types = self.lattice_constant_generator.get_lattice_constant_types()
		# Set up lattice_data.txt for writing the energies of lattice constants into. 
		energies_vs_lattice_constants = self.setup_energies_vs_lattice_constants()
		# Process the range of lattice constants and determine the minimum energy and lattice constant, as well as the projected lattice constant
		energies_vs_lattice_constants = self.get_data(energies_vs_lattice_constants)
		self.sort_data_file()
		energies_vs_lattice_constants, energy_vs_volumes = self.divide_up_energies_and_volumes(energies_vs_lattice_constants)
		self.minimum_energy, self.lowest_energy_lattice_constants = self.get_minimum_energy(energies_vs_lattice_constants)
		# Plot the data and give other information about the system
		self.plot_energy_vs_lattice_constants(energies_vs_lattice_constants)
		self.get_other_data_about_system(self.minimum_energy, self.lowest_energy_lattice_constants, energy_vs_volumes)

	############################################################################################################################

	def lattice_constant_parameters_checks(self,lattice_constant_parameters):
		"""
`		This method is designed to check that lattice_constant_parameters in in the correct format before contining
		"""
		
		if isinstance(lattice_constant_parameters,tuple) or isinstance(lattice_constant_parameters,list):
			pass
		elif isinstance(lattice_constant_parameters,dict):
			for lattice_constant_type, lattice_constant_parameter in lattice_constant_parameters.items():
				if isinstance(lattice_constant_parameter,tuple) or isinstance(lattice_constant_parameter,list):
					continue
		else:
			exit('error 3')

	############################################################################################################################

	def setup_energies_vs_lattice_constants(self):
		"""
		This method is designed to load information that has already been gather from previous LatticeFinder runs, or if not setting up the lattice_data_file that lattice constant data will be recorded to.
		"""
		if os.path.exists(self.lattice_data_file):
			energies_vs_lattice_constants = self.load_data()
		else:
			self.setup_initial_data_file()
			energies_vs_lattice_constants = {}
		return energies_vs_lattice_constants

	def load_data(self):
		"""
		Load data from previous LatticeFinder runs
		"""
		print('===============================================================')
		print('Loading data')
		energies_vs_lattice_constants = {}
		leading_data = False
		with open(self.lattice_data_file,'r') as lattice_data_FILE:
			for _ in range(12):
				lattice_data_FILE.readline()
			for line in lattice_data_FILE:
				lattice_constants, energy_per_atom = line.rstrip('\n').split(':')
				lattice_constants = eval(lattice_constants)
				if len(self.lattice_constant_types) == 1:
					energy_per_atom, volume = energy_per_atom.rstrip('\n').split('(')
					volume = float(volume.replace(')',''))
				energy_per_atom = float(energy_per_atom)
				if len(self.lattice_constant_types) == 1:
					energies_vs_lattice_constants[lattice_constants] = (energy_per_atom,volume)
				else:
					energies_vs_lattice_constants[lattice_constants] = energy_per_atom
		print('Finished Loading data')
		print('===============================================================')
		return energies_vs_lattice_constants

	def setup_initial_data_file(self):
		"""
		Set up the lattice_data_file file that data will be written to.
		"""
		with open(self.lattice_data_file,'w') as lattice_data_FILE:
			lattice_data_FILE.write('Symbol: '+str(self.symbol)+'\n')
			lattice_data_FILE.write('Lattice_type: '+str(self.lattice_type_name)+'\n')
			lattice_data_FILE.write('calculator: '+str(self.calculator)+'\n')
			lattice_data_FILE.write('size: '+str(self.size)+'\n')
			lattice_data_FILE.write('directions: '+str(self.directions)+'\n')
			lattice_data_FILE.write('miller: '+str(self.miller)+'\n')
			lattice_data_FILE.write('Lattice Constant Parameters: '+str(self.lattice_constant_types)+'\n')
			lattice_data_FILE.write('\n')
			lattice_data_FILE.write('Lattice Constant Results: \n')
			lattice_data_FILE.write('\n')
			to_write_to = 'Lattice constant ('+str(','.join(self.lattice_constant_types))+')/Ang: Energy/eV'
			if len(self.lattice_constant_types) == 1:
				to_write_to += ' (Volume (Ang^3 per atom))'
			to_write_to += '\n'
			lattice_data_FILE.write(to_write_to)
			lattice_data_FILE.write('----------------------------------------------\n')

	############################################################################################################################
	
	def get_data(self, energies_vs_lattice_constants):
		"""

		"""
		print('===============================================================')
		print('Getting data')
		if self.calculator == 'vasp':
			energies_vs_lattice_constants = get_energies_across_lattice_constants_VASP(self.lattice_type,self.symbol,self.lattice_constant_generator,self.size,directions=self.directions,miller=self.miller,lattice_data_file=self.lattice_data_file,vasp_inputs=self.VASP_Input_files,folder_name=self.VASP_Systems_Folder,slurm_information=self.slurm_information,force_rewrite=False,lattice_type_name=self.lattice_type_name,make_packets=self.make_packets)
		elif self.calculator == 'manual mode':
			energies_vs_lattice_constants = get_energies_across_lattice_constants_in_Manual_Mode(self.lattice_type,self.symbol,self.lattice_constant_generator,self.size,self.directions,self.miller,self.Manual_Mode_Cluster_Folder)
		else:
			energies_vs_lattice_constants = get_energies_across_lattice_constants_ASE(self.lattice_type,self.symbol,self.lattice_constant_generator,self.lattice_constant_types,self.size,self.directions,self.miller,self.calculator,self.no_of_cpus,self.lattice_data_file,energies_vs_lattice_constants)
		print('Finished getting data')
		print('===============================================================')
		return energies_vs_lattice_constants

	def sort_data_file(self):
		"""
		If you are using multiple processes, the ordering or lattice constants in the lattice_data.txt can become muddled up
		# This method reorders lines
		"""
		new_suffix = '.new'
		with open(self.lattice_data_file+new_suffix,'w') as lattice_data_FILE_NEW:
			lines = []
			with open(self.lattice_data_file,'r') as lattice_data_FILE_OLD:
				for _ in range(12):
					lattice_data_FILE_NEW.write(lattice_data_FILE_OLD.readline())
				for line in lattice_data_FILE_OLD:
					lattice_constants = eval(line.rstrip().split(':')[0])
					lines.append((lattice_constants,line))
			lines.sort()
			for lattice_constants, line in lines:
				lattice_data_FILE_NEW.write(line)
		os.remove(self.lattice_data_file)
		os.rename(self.lattice_data_file+new_suffix,self.lattice_data_file)

	def divide_up_energies_and_volumes(self, energies_vs_lattice_constants):

		if len(self.lattice_constant_types) == 1:
			new_energies_vs_lattice_constants = {}
			energy_vs_volumes = [[], []]
			energies_vs_lattice_constants = sorted(energies_vs_lattice_constants.items())
			total_no_of_enteries = len(energies_vs_lattice_constants)
			for _ in range(total_no_of_enteries):
				lattice_constant, (energy, volume) = energies_vs_lattice_constants.pop(0)
				new_energies_vs_lattice_constants[lattice_constant] = energy
				energy_vs_volumes[0].append(volume)
				energy_vs_volumes[1].append(energy)
			energies_vs_lattice_constants = new_energies_vs_lattice_constants
		else:
			energy_vs_volumes = None
		return energies_vs_lattice_constants, energy_vs_volumes


	def get_minimum_energy(self,energies_vs_lattice_constants):
		"""

		"""
		lowest_energy_per_atoms = float('inf')
		all_lattice_constants = []
		for lattice_constants, energy_per_atoms in energies_vs_lattice_constants.items():
			if energy_per_atoms <= lowest_energy_per_atoms:
				#if not isinstance(lattice_constants,dict):
				#	lattice_constants = [lattice_constants]
				if energy_per_atoms < lowest_energy_per_atoms:
					lowest_energy_per_atoms = energy_per_atoms
					all_lattice_constants = [lattice_constants]
				elif energy_per_atoms == lowest_energy_per_atoms:
					all_lattice_constants.append(lattice_constants)
		return lowest_energy_per_atoms, all_lattice_constants

	############################################################################################################################

	def plot_energy_vs_lattice_constants(self, energies_vs_lattice_constants):
		"""

		"""
		if len(self.lattice_constant_types) == 1:
			plot_energy_vs_lattice_constants_1D(energies_vs_lattice_constants, self.limits, self.plotting_limits, self.minimum_energy, self.lowest_energy_lattice_constants)
		if len(self.lattice_constant_types) == 2:
			plot_energy_vs_lattice_constants_2D(energies_vs_lattice_constants, self.limits, self.plotting_limits, self.minimum_energy, self.lowest_energy_lattice_constants)

	def get_other_data_about_system(self,minimum_energy,lowest_energy_lattice_constants,energy_vs_volumes=None):
		"""

		"""
		if isinstance(lowest_energy_lattice_constants[0],float):
			#latticeconstants = [{'c': lowest_energy_a_lattice_constants} for lowest_energy_a_lattice_constants in lowest_energy_lattice_constants]
			latticeconstants = lowest_energy_lattice_constants
		else:
			latticeconstants = [{key: value for key, value in zip(self.lattice_constant_types,lowest_energy_a_lattice_constants)} for lowest_energy_a_lattice_constants in lowest_energy_lattice_constants]
		if self.calculator == 'vasp':
			bulk_system = get_system_from_VASP(latticeconstants,folder_name=self.VASP_Systems_Folder)
		elif self.calculator == 'manual mode':
			bulk_system = get_system_from_Manual_Mode(latticeconstants,self.Manual_Mode_Cluster_Folder)
		else:
			bulk_system = get_system_from_ASE(self.lattice_type,self.symbol,latticeconstants,self.size,self.directions,self.miller,self.calculator)
		bulk_system = bulk_system[0]
		energy = bulk_system.get_potential_energy()
		no_of_atoms = len(bulk_system)
		energy_per_atom = energy/float(no_of_atoms)

		Volume = bulk_system.get_volume()
		Volume_per_atom = round(float(Volume)/float(no_of_atoms),9)
		#stress_tensor = bulk_system.get_stress(voigt=False)

		if len(self.lattice_constant_types) == 1:
			from ase.units import kJ
			from ase.eos import EquationOfState
			volumes = energy_vs_volumes[0]
			energies = energy_vs_volumes[1]
			eos = EquationOfState(volumes, energies, eos='sj')
			v0, e0, BB = eos.fit()
			bulk_modulus = (BB / kJ) * 1.0e24
			name_eos = 'Equation_of_State_Plot'
			# Save figure
			eos.plot()
			plt.savefig(name_eos+'.png')
			plt.savefig(name_eos+'.eps')
			plt.savefig(name_eos+'.svg')
			plt.cla(); plt.clf()

		with open('results_file.txt','w') as results_fileTXT:
			results_fileTXT.write('Symbol: '+str(self.symbol)+'\n')
			results_fileTXT.write('Lattice_type: '+str(self.lattice_type_name)+'\n')
			results_fileTXT.write('calculator: '+str(self.calculator)+'\n')
			results_fileTXT.write('size: '+str(self.size)+'\n')
			results_fileTXT.write('directions: '+str(self.directions)+'\n')
			results_fileTXT.write('miller: '+str(self.miller)+'\n')
			results_fileTXT.write('Lattice Constant Parameters: '+str(self.lattice_constant_types)+'\n')
			results_fileTXT.write('\n')
			results_fileTXT.write('Properties of System: \n')	
			results_fileTXT.write('\n')
			results_fileTXT.write('Total energy: '+str(energy)+' eV\n')
			results_fileTXT.write('No. of atoms: '+str(no_of_atoms)+' Atoms (Note the number of atoms along each natural direction of the bulk is '+str(self.size)+')\n')
			results_fileTXT.write('Cohesive energy: '+str(energy_per_atom)+' eV/Atom\n')
			results_fileTXT.write('\n')
			results_fileTXT.write('Total Volume: '+str(Volume)+' Angstroms^3\n')
			results_fileTXT.write('Volume per atom: '+str(Volume_per_atom)+' Angstroms^3/Atom\n')
			if not (self.calculator == 'vasp' or self.calculator == 'manual mode'):
				results_fileTXT.write('\n')
				stress_tensor = bulk_system.get_stress(voigt=False)
				results_fileTXT.write('Stress tensor:\n'+str(stress_tensor)+'\n')
			if len(self.lattice_constant_types) == 1:
				results_fileTXT.write('\n')
				results_fileTXT.write('Bulk Modulus: '+str(bulk_modulus)+' GPa\n')

	############################################################################################################################


import os
from copy import deepcopy
from ase.io import read
from shutil import copyfile
from LatticeFinder.LatticeFinder.examining_lattice_constant_methods_with_Manual_Mode import save_cluster_to_folder
from LatticeFinder.LatticeFinder.Create_submitSL_slurm_Main import make_submitSL, make_submitSL_packets_for_latticeFinder
from LatticeFinder.LatticeFinder.examining_lattice_constant_methods_with_ASE import save_datum_to_file, get_volume_per_atom

def get_energies_across_lattice_constants_VASP(lattice_type,symbol,lattice_constant_generator,size,directions=None,miller=None,lattice_data_file=None,vasp_inputs='VASP_Files',folder_name='VASP_Clusters',slurm_information={},force_rewrite=False,lattice_type_name=None,make_packets='packet'):
	"""

	"""
	if force_rewrite or make_new_cell_folders(folder_name,lattice_constant_generator):
		print('Not getting data, instead making clusters for running in VASP.')
		make_VASP_folders(lattice_type,symbol,lattice_constant_generator,size,directions,miller,vasp_inputs,folder_name,slurm_information,force_rewrite,make_packets)
		exit()
	else:
		slurm_information = get_VASP_energies(lattice_constant_generator,lattice_data_file,folder_name,lattice_type_name)
	return slurm_information

def make_new_cell_folders(folder_name,lattice_constant_generator):
	"""

	"""
	if not os.path.exists(folder_name):
		return True
	lattice_constant_generator.reset()
	all_lattice_constants = list(lattice_constant_generator)
	for index in range(len(all_lattice_constants)):
		latticeconstants = all_lattice_constants[index]
		if isinstance(latticeconstants,dict):
			name = '_'.join([str(key)+'_'+str(value) for key, value in latticeconstants.items()])
		else:
			name = 'c_'+str(latticeconstants)
		all_lattice_constants[index] = name
	all_folders = tuple(directory for directory in os.listdir(folder_name) if os.path.isdir(folder_name+'/'+directory))

	for name in all_lattice_constants:
		if not (name in all_folders):
			lattice_constant_generator.reset()
			return True
	lattice_constant_generator.reset()
	return False

def get_VASP_energies(lattice_constant_generator,lattice_data_file=None,folder_name='VASP_Clusters',lattice_type_name=None):
	"""

	"""
	check_lattice_data_file(lattice_data_file)
	slurm_information = {}
	clusters_not_calculated = []
	lattice_constant_generator.reset()
	for latticeconstants in lattice_constant_generator:
		if isinstance(latticeconstants,dict):
			name = '_'.join([str(key)+'_'+str(value) for key, value in latticeconstants.items()])
		else:
			name = 'c_'+str(latticeconstants)
		energy_per_atom, bulk_system = get_VASP_energy_per_atom(folder_name+'/'+name)
		if not isinstance(energy_per_atom,float):
			clusters_not_calculated.append((latticeconstants,folder_name+'/'+name))
			continue
		if isinstance(latticeconstants,dict):
			latticeconstants = tuple(lc_value for lc_key, lc_value in sorted(latticeconstants.items()))
			save_datum_to_file(lattice_data_file,latticeconstants,energy_per_atom, volume=None)
			slurm_information[latticeconstants] = energy_per_atom
		else:
			volume = get_volume_per_atom(bulk_system)
			save_datum_to_file(lattice_data_file,latticeconstants,energy_per_atom, volume=volume)
			slurm_information[latticeconstants] = (energy_per_atom,volume)
	if len(clusters_not_calculated):
		print('=======================================================================================')
		print('Error: The following data has not completed and need to be obtained before continuing.')
		for latticeconstants, path_to in clusters_not_calculated:
			print('Lattice Constant: '+str(latticeconstants)+ ' Angstrom ('+str(path_to)+')')
		lc_types = lattice_constant_generator.get_lattice_constant_types()
		if len(lc_types) == 1:
			print('As '+lattice_type_name+' only has one lattice constant, this is not an issue if you do not suspect this lattice constant to be the optimal lattice constant.')
			print('Will continue on, ONLY A WARNING HERE.')
		else:
			print('This program will finish without completing.')
			print('=======================================================================================')
			exit()
	lattice_constant_generator.reset()
	return slurm_information

def check_lattice_data_file(lattice_data_file):
	suffix = '.new'
	with open(lattice_data_file,'r') as lattice_data_file_old:
		with open(lattice_data_file+suffix,'w') as lattice_data_file_new:
			for _ in range(12):
				lattice_data_file_new.write(lattice_data_file_old.readline())
	os.remove(lattice_data_file)
	os.rename(lattice_data_file+suffix,lattice_data_file)

def get_VASP_energy_per_atom(folder):
	try:
		system = read(folder+'/OUTCAR')
		energy = system.get_potential_energy()
		no_of_atoms = float(len(system))
		energy_per_atom = energy/no_of_atoms
	except Exception:
		energy_per_atom, system = None, None
	return energy_per_atom, system


def make_VASP_folders(lattice_type,symbol,lattice_constant_generator,size,directions=None,miller=None,vasp_inputs='VASP_Files',folder_name='VASP_Clusters',slurm_information={},force_rewrite=False,make_packets='packets'):
	if not os.path.exists(folder_name):
		os.makedirs(folder_name)
	if not force_rewrite:
		all_folders = tuple(directory for directory in os.listdir(folder_name) if os.path.isdir(folder_name+'/'+directory))
	check_VASP_files(vasp_inputs)
	if make_packets == 'packets':
		all_directories = []
	for latticeconstants in lattice_constant_generator:
		if isinstance(latticeconstants,dict):
			name = '_'.join([str(key)+'_'+str(value) for key, value in latticeconstants.items()])
		else:
			name = 'c_'+str(latticeconstants)
		if not force_rewrite and (name in all_folders):
			continue
		bulk_system = lattice_type(symbol=symbol, latticeconstant=latticeconstants, size=size) #, directions=directions, miller=miller)
		bulk_system.set_pbc(True)
		save_cluster_to_folder(folder_name,name,'','vasp',bulk_system)
		copy_VASP_files(folder_name+'/'+name,vasp_inputs,slurm_information,make_packets)
		if make_packets == 'packets':
			all_directories.append(name)
	if make_packets == 'packets':
		make_packet_submitSL_files(folder_name,all_directories,slurm_information)

def check_VASP_files(vasp_files_folder):
	if not os.path.exists(vasp_files_folder):
		print('Error in copying VASP files to cluster folder')
		print('There is no folder called "VASP_Files" in your working directory.')
		print('This folder is where you should place your VASP files in for DFT local optimisations.')
		print('Make this folder and place your "POTCAR", INCAR", "KPOINTS" files for VASP local optimisations')
		print('This program will exit without completing.')
		exit()
	VASP_Files_files = os.listdir(vasp_files_folder)
	have_POTCAR  = 'POTCAR'  in VASP_Files_files
	have_INCAR   = 'INCAR'   in VASP_Files_files
	have_KPOINTS = 'KPOINTS' in VASP_Files_files
	if not (have_POTCAR and have_INCAR and have_KPOINTS):
		print('Error in copying VASP files to cluster folder')
		print('You need the following files when you are performing VASP calculations:')
		print()
		print('\tPOTCAR:\t' +('You have this' if have_POTCAR  else 'You do not have this'))
		print('\tINCAR:\t'  +('You have this' if have_INCAR   else 'You do not have this'))
		print('\tKPOINTS:\t'+('You have this' if have_KPOINTS else 'You do not have this'))
		print()
		print('Check this out. This program will now end without completing.')
		exit()

def copy_VASP_files(root,vasp_files_folder,slurm_information,packet_submit_sl_file):
	check_INCAR_Further(vasp_files_folder)
	print('Copying VASP files to '+root)
	for file in ['POTCAR', 'INCAR', 'KPOINTS']:
		copyfile(vasp_files_folder+'/'+file, root+'/'+file)
	if make_packets == 'packets':
		project = slurm_information['project']
		time = slurm_information['time']
		nodes = slurm_information['nodes']
		ntasks_per_node = slurm_information['ntasks_per_node']
		mem_per_cpu = slurm_information['mem-per-cpu']
		partition = slurm_information['partition']
		email = slurm_information['email']
		python_version = slurm_information['python_version']
		vasp_version = slurm_information['vasp_version']
		vasp_execution = slurm_information['vasp_execution']
		make_submitSL(root,project,time,nodes,ntasks_per_node,mem_per_cpu,partition=partition,email=email,python_version=python_version,vasp_version=vasp_version,vasp_execution=vasp_execution)

def make_packet_submitSL_files(root,all_directories,slurm_information):
	project = slurm_information['project']
	time = slurm_information['time']
	nodes = slurm_information['nodes']
	ntasks_per_node = slurm_information['ntasks_per_node']
	mem_per_cpu = slurm_information['mem-per-cpu']
	partition = slurm_information['partition']
	email = slurm_information['email']
	python_version = slurm_information['python_version']
	vasp_version = slurm_information['vasp_version']
	vasp_execution = slurm_information['vasp_execution']
	# other stuff
	number_of_vasp_calc_to_run_per_packet = slurm_information['Number of VASP calculations to run per packet']
	make_submitSL_packets_for_latticeFinder(number_of_vasp_calc_to_run_per_packet,all_directories,root,project,time,nodes,ntasks_per_node,mem_per_cpu,partition=partition,email=email,python_version=python_version,vasp_version=vasp_version,vasp_execution=vasp_execution)


def check_INCAR_Further(vasp_files_folder):
	"""

	"""
	with open(vasp_files_folder+'/INCAR') as INCAR:
		for line in INCAR:
			if 'NSW' in line:
				line = line.rstrip().lstrip()
				exec(line)
				if NSW == 0:
					break
				else:
					print('================================================')
					print('Error in LatticeFinder')
					print('You need to set NSW = 0 in your INCAR')
					print('NSW = '+str(NSW))
					print('Check this out. This program will finish without completing.')
					print('================================================')
					exit()
		else:
			print('================================================')
			print('Error in LatticeFinder')
			print('You need to set NSW = 0 in your INCAR')
			prtin('You do not have NSW in your INCAR')
			print('Check this out. This program will finish without completing.')
			print('================================================')
			exit()

def get_system_from_VASP(latticeconstants,folder_name):
	"""

	"""
	systems = []
	for a_latticeconstants in latticeconstants:
		if isinstance(a_latticeconstants,dict):
			name = '_'.join([str(key)+'_'+str(value) for key, value in a_latticeconstants.items()])
		else:
			name = 'c_'+str(a_latticeconstants)
		system = read(folder_name+'/'+name+'/OUTCAR')
		systems.append(system)
	return systems


